from flask_marshmallow import Marshmallow as _Marshmallow, Schema as _Schema


class Marshmallow(_Marshmallow):

    def __init__(self):
        super(Marshmallow, self).__init__()
        self.Schema = Schema


class Schema(_Schema):

    def handle_error(self, exc, data, **kwargs):
        for key in exc.messages.keys():
            exc.messages[key] = exc.messages[key][0]
        raise exc



from flask_restful.utils import http_status_message
from marshmallow import ValidationError

from frmf.db_helper import DBHelper
from frmf.response import abort

db_helper = DBHelper()


class Logic(object):
    """
    Remember to call super() function in __init__ when inheriting this class.
    """
    _manager = None
    _schema = None

    def __init__(self):
        self.error_messages = {}
        if self._schema is not None:
            self.schema = self._schema()
        if self._manager is not None:
            self.manager = self._manager()

    def list(self, query_args, *args, **kwargs):

        page = query_args.pop('page')
        results_per_page = query_args.pop('results_per_page')
        sort = query_args.pop('sort')
        data = self.manager.list(page=page, results_per_page=results_per_page, sort=sort, **query_args)
        if data:
            return True, self.schema.dump(data, many=True), 200
        else:
            return True, [], 204

    def retrieve(self, pk, *args, **kwargs):

        data = self.manager.retrieve(pk=pk)
        if data is None:
            return False, {'error': http_status_message(404)}, 404
        return True, self.schema.dump(data), 200

    def create(self, data, query_args, *args, **kwargs):
        with db_helper.auto_commit():
            data = self.validate(data)
            data = self.manager.create(data)
        return True, self.schema.dump(data), 201

    def update(self, pk, data, partial=False, *args, **kwargs):

        with db_helper.auto_commit():
            data = self.validate(data, partial=partial)
            data = self.manager.update(pk, data)
        return True, self.schema.dump(data), 200

    def delete(self, pk, *args, **kwargs):
        with db_helper.auto_commit():
            self.manager.delete(targets=pk)

        return True, None, 204

    def validate(self, data, schema=None, *args, **kwargs):

        try:
            if schema is None:
                data = self.schema.load(data, partial=kwargs.get('partial', False))
            else:
                data = schema().load(data, partial=kwargs.get('partial', False))
        except ValidationError as e:
            abort(400, e.normalized_messages())
        else:
            return data



from frmf.api_view import ApiView
from frmf.logic import Logic
from frmf.manager import Manager
from frmf.db_helper import DBHelper
from frmf.response import abort
from frmf.response import make_response


from frmf.db_helper import DBHelper
from frmf.response import abort

db_helper = DBHelper()


class Manager:
    _model = None
    _db = None

    def list(self, page=None, results_per_page=None, sort=None, *args, **kwargs):

        query = self._model.query
        if page and results_per_page:
            offset = (page - 1) * results_per_page
            query = query.limit(results_per_page).offset(offset)

        for key, value in kwargs.items():
            if isinstance(value, list):
                if key.endswith('_between'):
                    key = key.replace('_between', '')
                    query = query.filter(eval(f'self._model.{key}').between(value[0], value[1]))
                elif key.endswith('[]'):
                    key = key.replace('[]', '')
                    query = query.filter(self._model.__dict__[key].in_(value))
                else:
                    query = query.filter(self._model.__dict__[key].in_(value))
            else:
                if isinstance(value, bool):
                    query = query.filter_by(**{key: value})
                else:
                    if value:
                        if key.endswith('_in'):
                            query = query.filter(self._model[key].in_(value))

                        else:
                            query = query.filter_by(**{key: value})
        if sort:
            sort_list = sort.split(',')
            for criteria in sort_list:
                try:
                    sort_by, order = criteria.split(':')
                except ValueError:
                    sort_by = criteria
                    order = 'asc'
                if hasattr(self._model, sort_by):
                    if order == 'asc':
                        query = query.order_by(self._model.__dict__[sort_by])
                    elif order == 'desc':
                        query = query.order_by(self._model.__dict__[sort_by].desc())
        return query.all()

    def retrieve(self, pk=None, *args, **kwargs):
        query = self._model.query
        if pk:
            query = query.get(pk)
            return query
        elif kwargs:
            for key, value in kwargs.items():
                if isinstance(value, list):
                    key = key.replace('[]', '')
                    query = query.filter(self._model.__dict__[key].in_(value))
                else:
                    if isinstance(value, bool):
                        query = query.filter_by(**{key: value})
                    else:
                        if value:
                            query = query.filter_by(**{key: value})

            return query.first()
        else:
            raise abort(404)

    def create(self, data, *args, **kwargs):
        res = self._model(**data)
        db_helper.add(res)

        return res

    def update(self, pk, data, partial=False, *args, **kwargs):
        obj = self._model.query.get(pk)
        if partial and obj is None:
            data['id'] = pk
            res = self.create(data)
            return res
        elif obj is None:
            abort(404)

        for key, value in data.items():
            setattr(obj, key, value)

        return obj

    def delete(self, targets, multi=False, *args, **kwargs):

        if multi:
            if not isinstance(targets, list):
                raise TypeError()
            for target in targets:
                instance = self._model.query.get(target)
                if instance:
                    db_helper.delete(instance)
                    db_helper.flush()

        else:
            instance = self._model.query.get(targets)
            db_helper.delete(instance)

    def delete_by_instance(self, targets, multi=False, *args, **kwargs):

        if multi:
            if not isinstance(targets, list):
                raise TypeError()
            for target in targets:
                db_helper.delete(target)
        else:
            db_helper.delete(targets)



from contextlib import contextmanager

from frmf.response import abort


class _MetaSingleton(type):
    _instance = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instance:
            cls._instance[cls] = super(_MetaSingleton, cls).__call__(*args, **kwargs)
        return cls._instance[cls]


class _DBHelperConfig(object):
    def __init__(self, db_helper, db, **kwargs):
        self.db_helper = db_helper
        self.db = db
        self.configure_args = kwargs

    @property
    def metadata(self):
        return self.db.metadata


class DBHelper(metaclass=_MetaSingleton):

    def __init__(self, app=None, db=None, **kwargs):
        self.db = None
        if app is not None and db is not None:
            self.init_app(app, db)

    def init_app(self, app, db):
        self.db = db or self.db
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        app.extensions['db_helper'] = _DBHelperConfig(
            self, self.db)

    def add(self, instance):
        return self.db.session.add(instance)

    def delete(self, instance):
        return self.db.session.delete(instance)

    def flush(self):
        return self.db.session.flush()

    @contextmanager
    def auto_commit(self):
        try:
            yield self.db
            self.db.session.commit()
        except Exception as e:
            print(e)
            self.db.session.rollback()
            abort(500)



from flask import jsonify
from flask import make_response as _make_response
from flask import abort as _abort
from flask_restful.utils import http_status_message


def abort(status_code, messages=None):
    if messages is None:
        messages = {'error': http_status_message(status_code)}
    response = make_response(success=False, data=messages, status_code=status_code)
    _abort(response)


def make_response(success, data, status_code):
    if success:
        response = data
    else:

        response = {
            'messages': data,
        }

    return _make_response(jsonify(response), status_code)



from flask import request
from flask_restful import Resource, reqparse

from frmf.response import make_response


class ApiView(Resource):
    _logic = None
    decorators = []

    def __init__(self):
        self.parser = reqparse.RequestParser(bundle_errors=True)
        self.query_args = self.parser.parse_args()
        self.set_global_args()
        self.set_query_args()
        if self._logic is not None:
            self.logic = self._logic()

    def set_query_args(self):
        self.set_global_args()
        self.query_args = self.parser.parse_args()
        if request.method == 'GET':
            self.set_get_args()
        elif request.method == 'POST':
            self.set_post_args()
        elif request.method == 'PUT':
            self.set_put_args()
        elif request.method == 'PATCH':
            self.set_patch_args()
        elif request.method == 'DELETE':
            self.set_delete_args()
        self.query_args = self.parser.parse_args()

    def set_global_args(self):
        self.add_union_argument(arg1='page', arg2='results_per_page', type1=int, type2=int, location='args')
        self.parser.add_argument('sort', type=str, required=False, location='args')

    def add_mutex_argument(self, arg1, arg2, type1, type2, location):
        """
        arg1與arg2互斥
        :param arg1:
        :param arg2:
        :param type1:
        :param type2:
        :param location:
        :return:
        """
        self.parser.add_argument(arg1, type=type1, required=False, location=location)
        self.parser.add_argument(arg2, type=type2, required=False, location=location)
        self.query_args = self.parser.parse_args()
        if self.query_args[arg2]:
            self.parser.remove_argument(arg1)
        elif self.query_args[arg1]:
            self.parser.remove_argument(arg2)

        self.query_args = self.parser.parse_args()

    def add_union_argument(self, arg1, arg2, type1, type2, location):
        """
        arg1 與 arg2 必須同時被設定
        :param arg1:
        :param arg2:
        :param type1:
        :param type2:
        :param location:
        :return:
        """
        self.parser.add_argument(arg1, type=type1, required=False, location=location)
        self.parser.add_argument(arg2, type=type2, required=False, location=location)
        self.query_args = self.parser.parse_args()
        if self.query_args[arg2]:
            self.parser.replace_argument(arg1, type=type1, required=True, location=location)
        elif self.query_args[arg1]:
            self.parser.replace_argument(arg2, type=type2, required=True, location=location)

    def add_filter_argument(self, name, type, required, location):
        """
        一次加入單數與複數參數
        :param name:
        :param type:
        :param required:
        :param location:
        :return:
        """
        multiple_name = f'{name}[]'
        self.parser.add_argument(name, type=type, required=required, location=location)
        self.parser.add_argument(multiple_name, type=type, required=required, action='append', location=location)
        self.query_args = self.parser.parse_args()
        if self.query_args[name]:
            self.parser.remove_argument(multiple_name)
        elif self.query_args[multiple_name]:
            self.parser.remove_argument(name)

    def set_get_args(self):
        pass

    def get(self, pk=None):
        if pk is None:
            status, data, status_code = self.logic.list(query_args=self.query_args)
            return make_response(status, data, status_code)
        else:
            status, data, status_code = self.logic.retrieve(pk, query_args=self.query_args)
            return make_response(status, data, status_code)

    def set_post_args(self):
        pass

    def post(self):
        request_data = request.get_json(force=True)
        status, data, status_code = self.logic.create(data=request_data, query_args=self.query_args)
        return make_response(status, data, status_code)

    def set_put_args(self):
        pass

    def put(self, pk):
        request_data = request.get_json(force=True)
        status, data, status_code = self.logic.update(pk=pk, data=request_data, query_args=self.query_args)
        return make_response(status, data, status_code)

    def set_patch_args(self):
        pass

    def patch(self, pk):
        request_data = request.get_json(force=True)
        status, data, status_code = self.logic.update(pk=pk, data=request_data, partial=True,
                                                      query_args=self.query_args)
        return make_response(status, data, status_code)

    def set_delete_args(self):
        pass

    def delete(self, pk):
        status, data, status_code = self.logic.delete(pk=pk)
        return make_response(status, data, status_code)
